##################################################
##################################################
# Submodule for turning the console into a canvas #
# CGE = Console Game Engine #
##################################################
##################################################

##################################################
# Imports #
##################################################
from .colorain import *
Stx = StyledText


##################################################
# Block Elements #
##################################################
BLOCK_FULL = chr(9608)
BLOCK_THREEQTR = chr(9619)
BLOCK_HALF = chr(9618)
BLOCK_QTR = chr(9817)

BLOCKS = [BLOCK_FULL, BLOCK_THREEQTR, BLOCK_HALF, BLOCK_QTR]

##################################################
# Move the cursor relative to current position #
##################################################
def move_cursor(line, col):
    base = "\033["

    linedir = "A" if line < 0 else "B"
    coldir = "C" if col > 0 else "D"
    
    linecode = f"{base}{abs(line)}{linedir}"
    colcode = f"{base}{abs(col)}{coldir}"
    sys.stdout.write(linecode)
    sys.stdout.write(colcode) 


##################################################
# Get a one-character input w/o echoing #
##################################################
def getChar():
    try:
        # for Windows-based systems
        import msvcrt # If successful, we are on Windows
        return msvcrt.getch()

    except ImportError:
        # for POSIX-based systems (with termios & tty support)
        import tty, sys, termios  # raises ImportError if unsupported

        fd = sys.stdin.fileno()
        oldSettings = termios.tcgetattr(fd)

        try:
            tty.setcbreak(fd)
            answer = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, oldSettings)

        return answer


##################################################
# 2D coordinate #
##################################################
class Coord2D:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    
    def __repr__(self):
        return f"{(self.x, self.y)}"
    
    def __add__(self, coord):
        return Coord2D(self.x + coord.x, self.y + coord.y)
    
    def __sub__(self, coord):
        return Coord2D(self.x - coord.x, self.y - coord.y)
    
    def __eq__(self, coord):
        return self.x == coord.x and self.y == coord.y


##################################################
# 2D scene (i.e., 2D canvas) #
##################################################
class Scene2D:
    def __init__(self, width, height, colour = None):
        self.width = width
        self.height = height
        self.scene = [[BLOCKS[0]][:]*2*self.width for i in range(self.height)]
    
    def render(self, reset=True):
        sys.stdout.write('\n'.join([''.join(ln) for ln in self.scene]))
        if reset:
            move_cursor(-self.height+1, -2*self.width)
        else: print()

    def edit_pixel(self, pixel_loc, newval):
        if isinstance(pixel_loc, Coord2D):
            x, y = pixel_loc.x, pixel_loc.y
        elif isinstance(pixel_loc, tuple):
            x, y = pixel_loc[0], pixel_loc[1]
        self.scene[y][2*x] = newval
        self.scene[y][2*x+1] = newval
    
    def reset_pixel(self, pixel_loc):
        if isinstance(pixel_loc, Coord2D):
            x, y = pixel_loc.x, pixel_loc.y
        elif isinstance(pixel_loc, tuple):
            x, y = pixel_loc[0], pixel_loc[1]
        self.edit_pixel(x, y, BLOCKS[0])

    def paint_pixel(self, pixel_loc, colour):
        if isinstance(pixel_loc, Coord2D):
            x, y = pixel_loc.x, pixel_loc.y
        elif isinstance(pixel_loc, tuple):
            x, y = pixel_loc[0], pixel_loc[1]
        self.edit_pixel(x, y, Stx(f"<{fgtokens[colour]}>{BLOCKS[0]}</>").parsed)


class Sprite:
    def __init__(self, scene, positions):
        self.scene = scene
        self.positions = positions
        self.draw()

    def draw(self):
        self.scene.edit_pixel(self.positions[0], Stx(f"<f=lg;b=b>{BLOCKS[2]}</>").parsed)
        for pos in self.positions[1:]:
            self.scene.edit_pixel(pos, Stx(f"<f=lr;b=y>{BLOCKS[2]}</>").parsed)
    
    def erase(self):
        for pos in self.positions:
            self.scene.reset_pixel(pos)

    def translate(self, x, y):
        self.erase()
        for pos in self.positions:
            pos.x += x
            pos.y += y
        self.draw()


from .colours import *
from .errors import *
import re

##################################################
# Core styling class: StyledText. Every object in
# this library is a StyledText object or inherits
# from it.
##################################################
class StyledText:
    def __init__(self, text):
        if isinstance(text, str):
            self.text = text
        elif isinstance(text, StyledText):
            self.text = text.text
        else:
            raise InvalidInputError(str(type(text)))

        self.parsed = self.parse_color()
   
    def parse_color(self):
        return self.parse_tokens('color')
   
    def raw_text(self):
        return self.parse_tokens('raw')
   
    def parse_tokens(self, replacer):
        parsed_text = self.text
        
        # Defining token pattern
        TAG_PTN = r'<(B|I|U|/|(?:f=.{0,2})|(?:b=.{0,2}))' +\
                    r'(?:;\s*(B|I|U|(?:f=.{0,2})|(?:b=.{0,2})))?'*4 + '>'
        
        tags = re.findall(TAG_PTN, parsed_text)
        if len(tags):
            if self.text.count('</>')==0:
                raise MissingEndTagError

            for tag in tags:
                mx_replacer = parse_token(tag) if replacer=='color' else ''
                parsed_text = parsed_text.replace(f"<{';'.join(tag).strip(';')}>", mx_replacer)
        
        return parsed_text
   
    def add_fg_color(self, color):
        self.text = f'<{fgtokens[color]}>{self.raw_text()}<{fgtokens["end"]}>'
        self.parsed = self.parse_color()
    
    def add_bg_color(self, color):
        self.text = f'<{bgtokens[color]}>{self.raw_text()}<{fgtokens["end"]}>'
        self.parsed = self.parse_color()
    
    def __repr__(self):
        return self.parsed
    
    def __add__(self, clrtxt):
        if 'str' in str(type(clrtxt)):
            return StyledText(self.text+clrtxt)
        if 'colorain' in str(type(clrtxt)):
            return StyledText(self.text+clrtxt.text)


##################################################
# Foreground colouring: wrapper classes #
##################################################
class FGColor(StyledText):
    def __init__(self, text, color):
        super().__init__(text)
        self.add_fg_color(color)

class FGYellow(FGColor):
    def __init__(self, text):
        super().__init__(text,'yellow')

class FGRed(FGColor):
    def __init__(self, text):
        super().__init__(text, 'red')

class FGGray(FGColor):
    def __init__(self, text):
        super().__init__(text, 'darkgray')

class FGCyan(FGColor):
    def __init__(self, text):
        super().__init__(text, 'cyan')

class FGOrange(FGColor):
    def __init__(self, text):
        super().__init__(text, 'orange')

class FGGreen(FGColor):
    def __init__(self, text):
        super().__init__(text, 'green')

class FGPurple(FGColor):
    def __init__(self, text):
        super().__init__(text, 'purple')

class FGBlue(FGColor):
    def __init__(self, text):
        super().__init__(text, 'blue')

class FGLtRed(FGColor):
    def __init__(self, text):
        super().__init__(text, 'lightred')

class FGLtGreen(FGColor):
    def __init__(self, text):
        super().__init__(text, 'lightgreen')

class FGLtCyan(FGColor):
    def __init__(self, text):
        super().__init__(text, 'lightcyan')

class FGLtPurple(FGColor):
    def __init__(self, text):
        super().__init__(text, 'lightpurple')

class FGLtBlue(FGColor):
    def __init__(self, text):
        super().__init__(text, 'lightblue')


##################################################
# Background colouring: wrapper classes #
##################################################
class BGColor(StyledText):
    def __init__(self, text, color):
        super().__init__(text)
        self.add_bg_color(color)

class BGYellow(BGColor):
    def __init__(self, text):
        super().__init__(text, 'yellow')

class BGRed(BGColor):
    def __init__(self, text):
        super().__init__(text, 'red')

class BGGreen(BGColor):
    def __init__(self, text):
        super().__init__(text, 'green')

class BGGray(BGColor):
    def __init__(self, text):
        super().__init__(text, 'darkgray')

class BGCyan(BGColor):
    def __init__(self, text):
        super().__init__(text, 'cyan')

class BGBlue(BGColor):
    def __init__(self, text):
        super().__init__(text, 'blue')

class BGPurple(BGColor):
    def __init__(self, text):
        super().__init__(text, 'purple')


##################################################
# General formatting: wrapper classes #
##################################################
class FmtText(StyledText):
    def __init__(self, text, fmt):
        super().__init__(f"<{fmt}>{text}</>")

class Italic(FmtText):
    def __init__(self, text):
        super().__init__(text, 'I') 

class Bold(FmtText):
    def __init__(self, text):
        super().__init__(text, 'B')

class Underline(FmtText):
    def __init__(self, text):
        super().__init__(text, 'U')


##################################################
# Helper function for viewing available styles #
##################################################
def styles_guide():
    print(StyledText("<f=y;b=r;U>Foreground:</>"))
    for fmt in FMT_CODE:
        print(StyledText(f"<{fmt}>{FMT_NAMES[fmt]}</>: <<f=lr>{fmt}</>>"))
    for clr in fgtokens:
        token = fgtokens[clr]
        print(StyledText(f"<{token}>{clr}</>: <<f=lr>{token}</>>"))

    print(StyledText("<f=y;b=r;U>Background:</>"))
    for clr in bgtokens:
        token = bgtokens[clr]
        print(StyledText(f"<{token}>{clr}</>: <<f=lr>{token}</>>"))


##################################################
# Welcome message in interactive mode #
##################################################
import __main__ as main
if not hasattr(main, '__file__'):
    print(StyledText("<f=o>############################################################</>"))
    print(StyledText(f"<f=lc>Welcome to</> <f=lr>c<f=o>o<f=y>l<f=g>o<f=b>r<f=gr>a<f=lr>i<f=c>n </> <B>{VERSION}</>!"))
    print(StyledText("<f=lc>For getting help on styling, run <f=r>styles_guide()</>. <f=y>Have fun!</>"))
    print(StyledText("<f=o>############################################################</>"))
    print()


from .errors import *


##################################################
# version info #
##################################################
VERSION = "0.0.43"


##################################################
# Styling codes #
##################################################
CLR_CODE = {
    'black':0,
    'red':1,
    'green':2,
    'yellow':3,
    'blue':4,
    'purple':5,
    'cyan':6,
    'gray':7
}

FMT_CODE = {
    'B':1, # bold
    'I': 3, # italic
    'U':4 # underline
}

FMT_NAMES = {'B': 'bold', 'I': 'italic', 'U': 'underline'}


##################################################
# Block Elements #
##################################################
BLOCK_FULL = chr(9608)
BLOCK_THREEQTR = chr(9619)
BLOCK_HALF = chr(9618)
BLOCK_QTR = chr(9817)

BLOCKS = [BLOCK_FULL, BLOCK_THREEQTR, BLOCK_HALF, BLOCK_QTR]


##################################################
# Foreground and background #
##################################################
GRD_CODE = {'BG':4, 'FG':3}


##################################################
# Making the colour codes #
##################################################
def gcode(ground):
    return lambda clr: f"{GRD_CODE[ground]}{CLR_CODE[clr]}"


##################################################
# Making and parsing tokens for each colour #
##################################################
def make_tokens(ground, names):
    tokens = {}
    for key in names:
        if 'black'==key: 
            alias = 'k'
        elif 'gray'==key: 
            alias = 'gr'
        else:
            alias = key[0]
        if ground=='FG':
            if 'light' in key:
                if 'black' in key: alias+='k'
                elif 'gray' in key: alias+='gr'
                else:
                    alias += key[key.find('light')+len('light')]

        tokens[key] = f'{"f" if ground=="FG" else "b"}={alias}'
    
    return tokens


def parse_token(token):
    parsed_token = {}
    fmt_tags = ['B', 'I', 'U']

    for subtoken in token:
        if len(subtoken):
            if subtoken[0] in ['f','b']:
                abbrs = fg_abbrs if subtoken[0]=='f' else bg_abbrs
                prop_val = subtoken.split('=')[1].strip()
                if prop_val not in abbrs:
                    if prop_val:
                        raise InvalidValueError(subtoken[0], prop_val)
                    else:
                        raise MissingValueError(subtoken[0])
                if subtoken[0]=='f':
                    parsed_token['fg'] = fg_abbrs[prop_val]
                else:
                    parsed_token['bg'] = bg_abbrs[prop_val]

            elif subtoken[0] in fmt_tags:
                parsed_token[subtoken[0]] = True

            elif subtoken[0]=='/':
                parsed_token['end'] = True

            else:
                raise InvalidPropError(subtoken[0])

    return parsed_token_to_code(**parsed_token)


def parsed_token_to_code(bg = None, fg = None, B=None, I=None, U=None, end=None):
    code = '\033['

    if bg: code += f"{bgcodes[bg]};"
    if fg: code += f"{fgcodes[fg]};"
    if B: code += f"{FMT_CODE['B']};"
    if U: code += f"{FMT_CODE['U']};"
    if I: code += f"{FMT_CODE['I']};"
    if end: code += f"{fgcodes['end']};"

    code = code.strip(';')
    code += "m"

    return code


def make_clrcodes():
    fgcodes = {}
    bgcodes = {}
    
    make_fgcode = gcode('FG')
    make_bgcode = gcode('BG')
    
    for clr in CLR_CODE:
        fgcodes['orange' if clr=='yellow' else clr] = make_fgcode(clr)
        bgcodes[clr] = make_bgcode(clr)
    for clr in CLR_CODE:
        fgcodes[clr if clr=='yellow' else ('gray' if clr=='black' else 'light'+clr)] = f'1;{make_fgcode(clr)}'
    bgcodes['gray'] = make_bgcode('gray')
    
    return fgcodes, bgcodes


fgcodes, bgcodes = make_clrcodes()
fgcodes['end'], bgcodes['end'] = '0','0'

fgtokens = make_tokens('FG',fgcodes.keys())
bgtokens = make_tokens('BG', bgcodes.keys())

fg_abbrs = {fgtokens[k].split('=')[1]:k for k in fgtokens}
bg_abbrs = {bgtokens[k].split('=')[1]:k for k in bgtokens}

fgtokens['end'] = '/'
bgtokens['end'] = '/'


import sys
import time
from .colorain import *
from .console_game_engine import *
import random

Stx = StyledText

class Snake(Sprite):
    def __init__(self, head_init, scene):
        if isinstance(head_init, tuple):
            head_init = Coord2D(*head_init)
        self.positions =  [head_init, Coord2D(head_init.x-1, head_init.y)]
        super().__init__(scene, self.positions)

    def grow(self, direction):
        tail = self.positions[-1]
        if direction == 'VER':
            self.positions.append(Coord2D(tail.x, tail.y-1))
        elif direction == 'HOR':
            self.positions.append(Coord2D(tail.x-1, tail.y))
        elif direction in ['w', 'a', 's', 'd']:
            dx = -1 if direction=='d' else (1 if direction=='a' else 0)
            dy = -1 if direction=='s' else (1 if direction=='w' else 0)
            self.positions.append(Coord2D((tail.x + dx)%self.scene.width, (tail.y + dy)%self.scene.height))
        self.draw()

    def move(self, direction):
        self.erase()

        self.positions[0] += direction
        prevdir = direction

        for i in range(1,len(self.positions)):
            currpos = self.positions[i]
            currpos_old = Coord2D(currpos.x, currpos.y)
            prevpos = self.positions[i-1]
            
            if currpos.x == prevpos.x:
                if currpos.y > prevpos.y:
                    currpos.y -= 1 
                elif currpos.y < prevpos.y:
                    currpos.y += 1
                else:
                    pass
            elif currpos.y == prevpos.y:
                if currpos.x > prevpos.x:
                    currpos.x -= 1
                elif currpos.x < prevpos.x:
                    currpos.x += 1
                else:
                    pass
            else:
                if prevdir.x in [-1,1]:
                    if currpos.y < prevpos.y:
                        currpos.y += 1
                    elif currpos.y > prevpos.y:
                        currpos.y -= 1
                elif prevdir.y in [-1,1]:
                    if currpos.x < prevpos.x:
                        currpos.x += 1
                    elif currpos.x > prevpos.x:
                        currpos.x -= 1
            
            prevdir = currpos - currpos_old
        
        self.draw()


def spawn_apple(scene, snake):
    width = scene.width
    height = scene.height
    ok = False
    while not ok:
        ok = True
        apple_x = random.randint(0, width-1)
        apple_y = random.randint(0, height-1)
        apple_pos = Coord2D(apple_x, apple_y)

        for pos in snake.positions:
            if pos == apple_pos:
                ok = False
    scene.paint_pixel(apple_pos, 'lightred')
    
    return apple_pos

def update_scoreboard(prev_score, scoreboard):
    curr_score = prev_score + 1
    scoreboard_text = f"Score: {curr_score}"
    
    for i in range(len(scoreboard_text)):
        scoreboard.scene[0][i] = Stx(f"<f=b;b=gr>{scoreboard_text[i]}</>").parsed
    
    return curr_score, scoreboard


def snake_game(width, height, start_pos, fps = 10):
    gamesc = Scene2D(width, height)
    
    scoreboard = Scene2D(width, 1)
    score, scoreboard = update_scoreboard(-1, scoreboard)

    snake = Snake(Coord2D(*start_pos), gamesc)
    
    apple_pos = spawn_apple(gamesc, snake)
    
    while True:
        scoreboard.render(reset=False)
        gamesc.render()
        key = getChar()

        if key == 'w':
            snake.move(Coord2D(0,-1))
        elif key == 's':
            snake.move(Coord2D(0, 1))
        elif key == 'a':
            snake.move(Coord2D(-1,0))
        elif key == 'd':
            snake.move(Coord2D(1, 0))
        elif key == 'q':
            break
        
        if snake.positions[0] == apple_pos:
            score, scoreboard = update_scoreboard(score, scoreboard)
            snake.grow(key)
            apple_pos = spawn_apple(gamesc, snake)
        
        move_cursor(-1,-1)
        time.sleep(1/fps)
    
    return gamesc, snake


from setuptools import setup, find_packages

VERSION = '0.0.43' 
DESCRIPTION = 'A simple styling package for making console programmes great again.'
LONG_DESCRIPTION = open('README.md','r',encoding='utf-8').read()

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="colorain", 
        version=VERSION,
        author="Susmit Islam",
        author_email="<susmitislam31@gmail.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        url = 'https://github.com/susmit31/colorain',
        packages=find_packages(),
        install_requires=[], # add any additional packages that 
        # needs to be installed along with your package. Eg: 'caer'
        
        keywords=['colour','color','terminal']
)